from glob import glob
import os
import warnings

import markdown2
import toml


class BaseSourceFile(object):
    def __init__(self, parent, filename):
        slug, _ = os.path.splitext(filename)
        self._source = filename
        self._slug = slug
        self._parent = parent
        self._config = self._read_configuration(filename)
        self._add_markdown_attachments()
        self._convert_markdown()

    @property
    def slug(self):
        return self._slug

    @property
    def url(self):
        return '/%s' % self._slug

    def _read_configuration(self, filename):
        toml_file = '%s/%s' % (self._parent.source_dir, filename)
        with open(toml_file) as configuration:
            return toml.loads(configuration.read())

    def _add_markdown_attachments(self):
        pattern = '%s/%s.*.markdown' % (self._parent.source_dir, self.slug)
        for attachment in glob(pattern):
            key = attachment.split('.')[-2] + '_markdown'
            with open(attachment) as content:
                if key in self._config:
                    warnings.warn(
                        '"%s" in %s overriden by attachment file.' % (
                            key, self.slug))
                self._config[key] = content.read()

    def _convert_markdown(self):
        add = {}
        for key in self._config:
            if key[-9:] == '_markdown':
                dest = key[:-9]
                add[dest] = markdown2.markdown(self._config[key])
                if dest in self._config:
                    warnings.warn(
                        '"%s" in %s overriden by Markdown conversion.' % (
                            dest, self.slug))
        self._config.update(add)

    def __getattr__(self, key):
        if key == 'slug':
            return self.slug
        if key in self._config:
            return self._config[key]
        raise AttributeError

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __iter__(self):
        _keys = self._config.keys()
        _keys.append('slug')
        return iter(_keys)

    def __repr__(self):
        return '<flourish.TomlSourceFile object (%s)>' % self._source

    class DoesNotExist(Exception):
        pass


class MarkdownSourceFile(BaseSourceFile):
    def _read_configuration(self, filename):
        markdown_file = '%s/%s' % (self._parent.source_dir, filename)
        with open(markdown_file) as configuration:
            content = configuration.read()
        if content.startswith('---\n'):
            # 4 skips the starting `---\n`; 8 skips both
            end = content[4:].find('---\n')
            if end != -1:
                # we have a TOML block
                config = toml.loads(content[4:end+4])
                config['body_markdown'] = content[end+8:]
            else:
                raise RuntimeError(
                    '"%s" has no end marker for the frontmatter' % filename
                )
        else:
            config = {}
            config['body_markdown'] = content
        return config


class TomlSourceFile(BaseSourceFile):
    pass



from operator import attrgetter
import os
import warnings

from jinja2 import Environment, FileSystemLoader

from .source import MarkdownSourceFile, TomlSourceFile
from .url import URL


class Flourish(object):
    ARGS = [
        'source_dir',
        'templates_dir',
        'jinja',
    ]
    DATA = [
        '_filters',
        '_order_by',
        '_slice',
        '_source_files',
        '_urls',
    ]

    def __init__(
        self,
        source_dir='source',
        templates_dir='templates',
        output_dir='output',
        **kwargs
    ):
        self.source_dir = source_dir
        self.templates_dir = templates_dir
        self.output_dir = output_dir
        self.jinja = Environment(
            loader=FileSystemLoader(self.templates_dir),
            keep_trailing_newline=True,
        )

        self._filters = []
        self._order_by = []
        self._slice = None
        self._source_files = []
        self._urls = {}

        if not os.path.isdir(self.source_dir):
            raise AttributeError(
                'source_dir "%s" must exist' % self.source_dir)

        for _opt in self.DATA:
            if _opt in kwargs:
                if kwargs[_opt] is not None:
                    setattr(self, _opt, kwargs[_opt])

        if '_source_files' not in kwargs:
            self._add_sources()

    @property
    def sources(self):
        return self

    def all(self):
        """ Get all source documents. """
        return self

    def count(self):
        """ Get the count of all source documents. """
        _sources = 0
        for source in self.sources.all():
            _sources += 1
        return _sources

    def get(self, slug):
        """ Get a single source document by slug. """
        for source in self.sources.all():
            if source.slug == slug:
                return source
        raise TomlSourceFile.DoesNotExist

    def filter(self, **kwargs):
        _clone = self.clone()
        for _key, _value in kwargs.iteritems():
            _clone._filters.append((_key, _value))
        return _clone

    def order_by(self, *args):
        return self.clone(_order_by=args)

    def add_url(self, url, name, generator):
        _url_dict = {name: URL(self, url, name, generator)}
        self._urls.update(_url_dict)

    def resolve_url(self, name, **kwargs):
        return self._urls[name].resolve(**kwargs)

    def all_valid_filters_for_url(self, name):
        return self._urls[name].all_valid_filters()

    def generate_all_urls(self):
        for _entry in self._urls:
            url = self._urls[_entry]
            url.generator(self, url)

    def clone(self, **kwargs):
        for _option in self.ARGS + self.DATA:
            _value = getattr(self, _option)
            if type(_value) == list:
                kwargs.setdefault(_option, _value[:])
            else:
                kwargs.setdefault(_option, _value)
        return type(self)(**kwargs)

    def _add_sources(self):
        """ Find source documents and register them. """
        trim = len(self.source_dir) + 1
        for root, dirs, files in os.walk(self.source_dir):
            this_dir = root[trim:]
            for file in files:
                if len(this_dir):
                    file = '%s/%s' % (this_dir, file)
                if file.endswith('.toml'):
                    self._source_files.append(TomlSourceFile(self, file))
                elif file.endswith('.markdown') and len(file.split('.')) == 2:
                    self._source_files.append(MarkdownSourceFile(self, file))

    def __len__(self):
        return self.count()

    def __getitem__(self, item):
        if isinstance(item, slice):
            if (item.start < 0) or (item.stop < 0):
                raise ValueError('Cannot use negative indexes with Flourish')
            return self.clone(_slice=item)
        if item < 0:
            raise ValueError('Cannot use negative indexes with Flourish')
        _iterator = iter(self)
        try:
            for i in range(0, item+1):
                obj = next(_iterator)
        except StopIteration:
            raise IndexError(item)
        return obj
        return None

    def _get_filtered_sources(self):
        _sources = []
        for _source in self._source_files:
            _add = True
            if len(self._filters):
                for _filter in self._filters:
                    if '__' in _filter[0]:
                        _field, _operator = _filter[0].split('__', 2)
                    else:
                        _field, _operator = _filter[0], 'equal_or_inside'
                    _operation = OPERATORS.get(_operator)
                    if _operation is None:
                        # FIXME write test and raise more useful error
                        raise RuntimeError
                    if _operator == 'unset':
                        try:
                            getattr(_source, _field)
                            _add = False
                        except AttributeError:
                            pass
                    else:
                        try:
                            _result = _operation(
                                getattr(_source, _field), _filter[1]
                            )
                            if not _result:
                                _add = False
                        except AttributeError:
                            _add = False
            if _add:
                _sources.append(_source)
        if self._slice is not None:
            _sources = _sources.__getitem__(self._slice)
        return _sources

    def __iter__(self):
        _sources = self._get_filtered_sources()
        for _order in self._order_by:
            if _order[0] == '-':
                _rev = True
                _attr = _order[1:]
            else:
                _rev = False
                _attr = _order
            try:
                _sources = sorted(
                    _sources, key=attrgetter(_attr), reverse=_rev)
            except AttributeError:
                warnings.warn(
                    'sorting sources by "%s" failed: '
                    'not all sources have that attribute' % _order
                )

        return iter(_sources)

    def __repr__(self):
        return '<flourish.Flourish object (source=%s)' % self.source_dir


def _equal_to(value, test):
    return value == test


def _equal_or_inside(value, test):
    if type(value) == list:
        return test in value
    else:
        return value == test


def _less_than(value, test):
    return value < test


def _less_than_or_equal_to(value, test):
    return value <= test


def _greater_than(value, test):
    return value > test


def _greater_than_or_equal_to(value, test):
    return value >= test


def _contains(value, test):
    try:
        if type(value) == list:
            for _v in value:
                if test in _v:
                    return True
            return False
        else:
            return test in value
    except TypeError:
        return False


def _inside(value, test):
    return value in test


def _set(value, test):
    return value is not None


def _unset(value, test):
    return value is None


OPERATORS = {
    'equal_or_inside': _equal_or_inside,
    'equals': _equal_to,
    'lt': _less_than,
    'lte': _less_than_or_equal_to,
    'gt': _greater_than,
    'gte': _greater_than_or_equal_to,
    'contains': _contains,
    'in': _inside,
    'set': _set,
    'unset': _unset,
}



from setuptools import find_packages,setup

setup(
    name='flourish',
    version='0.1.0',

    description='Static website generator',
    # long_description='',
    url='https://github.com/norm/flourish',

    license='MIT',
    author='Mark Norman Francis',
    author_email='norm@201created.com',

    packages=find_packages(exclude=['tests']),

    install_requires=[
        'Jinja2',
        'markdown2',
        'toml',
    ],

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2',
        # 'Programming Language :: Python :: 3',

        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Text Processing :: Markup :: HTML',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
    ],
)



import os


class BaseGenerator(object):
    @classmethod
    def as_generator(cls):
        def generator(flourish, url):
            self = cls(flourish, url)
            return self.generate()

        return generator

    def __init__(self, flourish, url):
        self.flourish = flourish
        self.url = url

    def generate(self):
        for _filter in self.flourish.all_valid_filters_for_url(self.url.name):
            _objects = self.get_objects(_filter)
            _context = self.get_context_data(_objects)
            _template = self.get_template(_objects)
            _render = self.render_template(_template, _context)
            _path = self.url.resolve(**_filter)
            self.output_to_file(_path, _render)

    def get_template(self, objects):
        return self.flourish.jinja.get_template(
            self.get_template_name(objects)
        )

    def get_objects(self, filter):
        return self.flourish.sources.filter(**filter)

    def render_template(self, template, context_data):
        return template.render(context_data).encode('UTF-8')

    def output_to_file(self, path, render):
        _output_file = self.get_output_filename(path)
        with open(_output_file, 'w') as _output:
            _output.write(render)

    def get_template_name(self, objects):
        return self.template_name

    def get_output_filename(self, path):
        _destination = '%s%s' % (self.flourish.output_dir, path)
        if _destination.endswith('/'):
            _destination = _destination + 'index.html'
        if not _destination.endswith('.html'):
            _destination = _destination + '.html'

        _directory = os.path.dirname(_destination)
        if not os.path.isdir(_directory):
            os.makedirs(_directory)
        return _destination

    def get_context_data(self, objects):
        _context = {}
        _context['objects'] = objects
        return _context


class PageGenerator(BaseGenerator):
    def get_context_data(self, objects):
        _context = super(PageGenerator, self).get_context_data(objects)
        _context['page'] = objects[0]
        _context.update(objects[0]._config)
        return _context

    def get_template_name(self, objects):
        if 'template' in objects[0]:
            return objects[0]['template']
        if 'type' in objects[0]:
            return '%s.html' % objects[0]['type']
        return 'page.html'


class IndexGenerator(BaseGenerator):
    template_name = 'index.html'

    def get_context_data(self, objects):
        _context = super(IndexGenerator, self).get_context_data(objects)
        _context['pages'] = objects
        return _context



import re


class URL(object):
    def __init__(self, parent, path, name, generator):
        self.parent = parent
        self.path = path
        self.name = name
        self.generator = generator

    @property
    def arguments(self):
        _arguments = []
        for _segment in re.split('(#\w+)', self.path):
            if _segment.startswith('#'):
                _arguments.append(_segment[1:])
        return _arguments

    def resolve(self, **kwargs):
        _resolved = ''
        for _segment in re.split('(#\w+)', self.path):
            if _segment.startswith('#'):
                key = _segment[1:]
                if key in kwargs:
                    if kwargs[key] is None:
                        raise RuntimeError
                    _resolved = _resolved + kwargs[key]
                else:
                    raise KeyError
            else:
                _resolved = _resolved + _segment
        return _resolved

    def all_valid_filters(self):
        _valid_filters = []
        _args = self.arguments

        if len(_args) == 0:
            _valid_filters.append({})
        else:
            _filters = self.get_all_filters_for_arguments(_args)
            for _dict in _filters:
                _valid_filters.append(_dict)
        return _valid_filters

    def get_all_filters_for_arguments(self, keys, objects=None):
        if objects is None:
            objects = self.parent

        _first_key = keys[0]

        _values = set()
        for _source in objects:
            if _first_key in _source:
                if type(_source[_first_key]) == list:
                    for _value in _source[_first_key]:
                        _values.add(_value)
                else:
                    _values.add(_source[_first_key])

        _filters = []
        for _value in _values:
            _dict = {_first_key: _value}
            if len(keys) == 1:
                _filters.append(_dict)
            else:
                _sub_objects = objects.filter(**_dict)
                _sub_keys = self.get_all_filters_for_arguments(
                    keys[1:], _sub_objects)
                for _sub_key in _sub_keys:
                    _update = _dict.copy()
                    _update.update(_sub_key)
                    _filters.append(_update)

        # the sort is only so the tests can compare results easily
        return sorted(_filters)

    def __repr__(self):
        return '<flourish.URL object (%s: %s)>' % (self.name, self.path)
